#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from dataclasses import dataclass
from typing import Sequence
from typing import Tuple

import torch
from torch.distributions import Distribution
from torch.distributions import Normal

from disent.frameworks.vae._supervised__tvae import TripletVae
from disent.frameworks.vae._weaklysupervised__adavae import AdaVae
from disent.nn.loss.triplet import configured_dist_triplet
from disent.nn.loss.triplet import configured_triplet

log = logging.getLogger(__name__)


# ========================================================================= #
# Guided Ada Vae                                                            #
# ========================================================================= #


class AdaNegTripletVae(TripletVae):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    This is the supervised version of the Adaptive Triplet Loss for Variational Auto-Encoders
    - Triplets are usually ordered using the ground-truth distances over observations
    - Adaptive Triplet Loss is used to guide distance learning and encourage disentanglement
    """

    REQUIRED_OBS = 3

    @dataclass
    class cfg(TripletVae.cfg, AdaVae.cfg):
        # adavae
        ada_thresh_mode: str = "dist"  # only works for: adat_share_mask_mode == "posterior"
        # ada_tvae - loss
        # * this should be used with a schedule, slowly decrease from 1.0 down to 0.5 or less
        # * a similar schedule should also be used on `ada_thresh_ratio`, slowly increasing from 0.0 to 0.5
        adat_triplet_share_scale: float = 0.95
        # ada_tvae - averaging
        adat_share_mask_mode: str = "posterior"

    def hook_compute_ave_aug_loss(
        self,
        ds_posterior: Sequence[Normal],
        ds_prior: Sequence[Normal],
        zs_sampled: Sequence[torch.Tensor],
        xs_partial_recon: Sequence[torch.Tensor],
        xs_targ: Sequence[torch.Tensor],
    ):
        return self.estimate_ada_triplet_loss(
            ds_posterior=ds_posterior,
            cfg=self.cfg,
        )

    @staticmethod
    def estimate_ada_triplet_loss_from_zs(zs: Sequence[torch.Tensor], cfg: cfg):
        # compute shared masks, shared embeddings & averages over shared embeddings
        share_masks, share_logs = compute_triplet_shared_masks_from_zs(zs=zs, cfg=cfg)
        # compute loss
        ada_triplet_loss, ada_triplet_logs = AdaNegTripletVae.compute_ada_triplet_loss(
            share_masks=share_masks, zs=zs, cfg=cfg
        )
        # merge logs & return loss
        return ada_triplet_loss, {
            **ada_triplet_logs,
            **share_logs,
        }

    @staticmethod
    def estimate_ada_triplet_loss(ds_posterior: Sequence[Normal], cfg: cfg):
        # compute shared masks, shared embeddings & averages over shared embeddings
        share_masks, share_logs = compute_triplet_shared_masks(ds_posterior, cfg=cfg)
        # compute loss
        ada_triplet_loss, ada_triplet_logs = AdaNegTripletVae.compute_ada_triplet_loss(
            share_masks=share_masks, zs=(d.mean for d in ds_posterior), cfg=cfg
        )
        # merge logs & return loss
        return ada_triplet_loss, {
            **ada_triplet_logs,
            **share_logs,
        }

    @staticmethod
    def compute_ada_triplet_loss(share_masks, zs, cfg: cfg):
        # Normal Triplet Loss
        (a_z, p_z, n_z) = zs
        trip_loss = configured_triplet(a_z, p_z, n_z, cfg=cfg)

        # Soft Scaled Negative Triplet
        (ap_share_mask, an_share_mask, pn_share_mask) = share_masks
        triplet_hard_neg_ave_scaled = configured_dist_triplet(
            pos_delta=a_z - p_z,
            neg_delta=torch.where(an_share_mask, cfg.adat_triplet_share_scale * (a_z - n_z), (a_z - n_z)),
            cfg=cfg,
        )

        return triplet_hard_neg_ave_scaled, {
            "triplet": trip_loss,
            "triplet_chosen": triplet_hard_neg_ave_scaled,
        }


# ========================================================================= #
# AveAda-TVAE                                                               #
# ========================================================================= #


@dataclass
class AdaTripletVae_cfg(TripletVae.cfg, AdaVae.cfg):
    # adavae
    ada_thresh_mode: str = "dist"  # only works for: adat_share_mask_mode == "posterior"
    # ada_tvae - averaging
    adat_share_mask_mode: str = "posterior"


def compute_triplet_shared_masks_from_zs(zs: Sequence[torch.Tensor], cfg):
    """
    required config params:
    - cfg.ada_thresh_ratio:
    """
    a_z, p_z, n_z = zs
    # shared elements that need to be averaged, computed per pair in the batch.
    ap_share_mask = AdaVae.compute_shared_mask_from_zs(a_z, p_z, ratio=cfg.ada_thresh_ratio)
    an_share_mask = AdaVae.compute_shared_mask_from_zs(a_z, n_z, ratio=cfg.ada_thresh_ratio)
    pn_share_mask = AdaVae.compute_shared_mask_from_zs(p_z, n_z, ratio=cfg.ada_thresh_ratio)
    # return values
    share_masks = (ap_share_mask, an_share_mask, pn_share_mask)
    return share_masks, {
        "ap_shared": ap_share_mask.sum(dim=1).float().mean(),
        "an_shared": an_share_mask.sum(dim=1).float().mean(),
        "pn_shared": pn_share_mask.sum(dim=1).float().mean(),
    }


def compute_triplet_shared_masks(ds_posterior: Sequence[Distribution], cfg: AdaTripletVae_cfg):
    """
    required config params:
    - cfg.ada_thresh_ratio:
    - cfg.ada_thresh_mode: "kl", "symmetric_kl", "dist", "sampled_dist"
      : only applies if cfg.ada_share_mask_mode=="posterior"
    - cfg.adat_share_mask_mode: "posterior", "sample", "sample_each"
    """
    a_posterior, p_posterior, n_posterior = ds_posterior

    # shared elements that need to be averaged, computed per pair in the batch.
    if cfg.adat_share_mask_mode == "posterior":
        ap_share_mask = AdaVae.compute_shared_mask_from_posteriors(
            a_posterior, p_posterior, thresh_mode=cfg.ada_thresh_mode, ratio=cfg.ada_thresh_ratio
        )
        an_share_mask = AdaVae.compute_shared_mask_from_posteriors(
            a_posterior, n_posterior, thresh_mode=cfg.ada_thresh_mode, ratio=cfg.ada_thresh_ratio
        )
        pn_share_mask = AdaVae.compute_shared_mask_from_posteriors(
            p_posterior, n_posterior, thresh_mode=cfg.ada_thresh_mode, ratio=cfg.ada_thresh_ratio
        )
    elif cfg.adat_share_mask_mode == "sample":
        a_z_sample, p_z_sample, n_z_sample = a_posterior.rsample(), p_posterior.rsample(), n_posterior.rsample()
        ap_share_mask = AdaVae.compute_shared_mask_from_zs(a_z_sample, p_z_sample, ratio=cfg.ada_thresh_ratio)
        an_share_mask = AdaVae.compute_shared_mask_from_zs(a_z_sample, n_z_sample, ratio=cfg.ada_thresh_ratio)
        pn_share_mask = AdaVae.compute_shared_mask_from_zs(p_z_sample, n_z_sample, ratio=cfg.ada_thresh_ratio)
    elif cfg.adat_share_mask_mode == "sample_each":
        ap_share_mask = AdaVae.compute_shared_mask_from_zs(
            a_posterior.rsample(), p_posterior.rsample(), ratio=cfg.ada_thresh_ratio
        )
        an_share_mask = AdaVae.compute_shared_mask_from_zs(
            a_posterior.rsample(), n_posterior.rsample(), ratio=cfg.ada_thresh_ratio
        )
        pn_share_mask = AdaVae.compute_shared_mask_from_zs(
            p_posterior.rsample(), n_posterior.rsample(), ratio=cfg.ada_thresh_ratio
        )
    else:
        raise KeyError(f"Invalid cfg.adat_share_mask_mode={repr(cfg.adat_share_mask_mode)}")

    # return values
    share_masks = (ap_share_mask, an_share_mask, pn_share_mask)
    return share_masks, {
        "ap_shared": ap_share_mask.sum(dim=1).float().mean(),
        "an_shared": an_share_mask.sum(dim=1).float().mean(),
        "pn_shared": pn_share_mask.sum(dim=1).float().mean(),
    }


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import contextlib
import time
from concurrent.futures import ProcessPoolExecutor
from tempfile import NamedTemporaryFile

import h5py
import numpy as np
import pytest

from disent.dataset.data import Hdf5Dataset
from disent.dataset.data import XYObjectData
from disent.dataset.util.formats.hdf5 import H5Builder
from disent.dataset.util.formats.hdf5 import hdf5_resave_file
from disent.dataset.util.formats.hdf5 import hdf5_test_speed
from disent.util.function import wrapped_partial
from disent.util.inout.hashing import hash_file
from tests.util import no_stderr
from tests.util import no_stdout

# ========================================================================= #
# TESTS                                                                     #
# ========================================================================= #


# factors=(3, 3, 2, 3), len=54
TestXYObjectData = wrapped_partial(
    XYObjectData,
    grid_size=4,
    grid_spacing=1,
    min_square_size=1,
    max_square_size=2,
    square_size_spacing=1,
    palette="rgb_1",
)
_TEST_LEN = 54


def _iterate_over_data(data, indices):
    i = -1
    for i, idx in enumerate(indices):
        img = data[i]
    return i + 1


@contextlib.contextmanager
def create_temp_h5data(track_times=False, **h5py_dataset_kwargs):
    # generate data
    data = np.stack([img for img in TestXYObjectData()], axis=0)
    # create temp file
    with NamedTemporaryFile("r") as out_file:
        # create temp files
        with h5py.File(out_file.name, "w", libver="earliest") as file:
            file.create_dataset(
                name="data",
                shape=(_TEST_LEN, 4, 4, 3),
                dtype="uint8",
                data=data,
                track_times=track_times,
                **h5py_dataset_kwargs,
            )
        # return the data & file
        yield out_file.name, data


def test_hdf5_pickle_dataset():
    with create_temp_h5data() as (tmp_path, raw_data):
        # load the data
        # - ideally we want to test this with a pytorch
        #   DataLoader, but that is quite slow to initialise
        with Hdf5Dataset(tmp_path, "data") as data:
            indices = list(range(len(data)))
            # test locally
            assert _iterate_over_data(data=data, indices=indices) == _TEST_LEN
            # test multiprocessing
            executor = ProcessPoolExecutor(2)
            future_0 = executor.submit(_iterate_over_data, data=data, indices=indices[0::2])
            future_1 = executor.submit(_iterate_over_data, data=data, indices=indices[1::2])
            assert future_0.result() == _TEST_LEN // 2
            assert future_1.result() == _TEST_LEN // 2
            # test multiprocessing on invalid data
            with h5py.File(tmp_path, "r", swmr=True) as file:
                with pytest.raises(TypeError, match="h5py objects cannot be pickled"):
                    future_2 = executor.submit(_iterate_over_data, data=file["data"], indices=indices)
                    future_2.result()


@pytest.mark.parametrize(
    ["hash_mode", "target_hash"],
    [
        ("full", "a3b60a9e248b4b66bdbf4f87a78bf7cc"),
        ("fast", "a20d554d4912a39e7654b4dc98207490"),
    ],
)
def test_hdf5_determinism(hash_mode: str, target_hash: str):
    # check hashing a
    def make_and_hash(track_times=False, **h5py_dataset_kwargs):
        with create_temp_h5data(track_times=track_times, **h5py_dataset_kwargs) as (path_a, raw_data_a):
            a = hash_file(path_a, hash_type="md5", hash_mode=hash_mode, missing_ok=False)
        # track times only has a resolution of 1 second
        # TODO: this test is slow ~4.4 seconds of sleeping...
        time.sleep(1.1)
        # redo the same task
        with create_temp_h5data(track_times=track_times, **h5py_dataset_kwargs) as (path_b, raw_data_b):
            b = hash_file(path_b, hash_type="md5", hash_mode=hash_mode, missing_ok=False)
        return a, b

    # compute hashes
    deterministic_hash_a, deterministic_hash_b = make_and_hash(track_times=False)
    stochastic_hash_a, stochastic_hash_b = make_and_hash(track_times=True)
    # check hashes
    assert deterministic_hash_a == deterministic_hash_b
    assert stochastic_hash_a != stochastic_hash_b
    # check against target
    assert deterministic_hash_a == target_hash
    assert deterministic_hash_b == target_hash
    assert stochastic_hash_a != target_hash
    assert stochastic_hash_b != target_hash


@pytest.mark.parametrize(
    ["chunk_shape", "compression_lvl", "target_hash"],
    [
        ((1, 4, 4, 3), None, "0a60b3edc4e2d75f3d38de9f1a29700c"),
        ("batch", 4, "88801fba75bc6eb53ccfcc145ab8a4f5"),
    ],
)
def test_hdf5_resave_dataset(chunk_shape, compression_lvl, target_hash):
    with no_stdout(), no_stderr():
        with create_temp_h5data(chunks=(_TEST_LEN, 4, 4, 3)) as (inp_path, raw_data), NamedTemporaryFile(
            "r"
        ) as out_file, NamedTemporaryFile("r") as alt_file:
            out_path = out_file.name
            alt_path = alt_file.name
            # convert dataset
            hdf5_resave_file(
                inp_path=inp_path,
                out_path=out_path,
                dataset_name="data",
                chunk_size=(1, 4, 4, 3),
                compression="gzip" if (compression_lvl is not None) else None,
                compression_lvl=compression_lvl,
                batch_size=None,
                out_dtype=None,
                out_mutator=None,
                obs_shape=None,
                write_mode="w",
            )
            # check alt conversion
            with h5py.File(inp_path, "r") as inp, H5Builder(alt_path, mode="w") as builder:
                builder.add_dataset_from_array(
                    array=inp["data"],
                    name="data",
                    chunk_shape=chunk_shape,
                    compression_lvl=compression_lvl,
                    batch_size="auto",
                )
            # check datasets
            with h5py.File(inp_path, "r") as inp:
                assert np.all(inp["data"][...] == raw_data)
                assert inp["data"].chunks == (_TEST_LEN, 4, 4, 3)
            with h5py.File(out_path, "r") as out:
                assert np.all(out["data"][...] == raw_data)
                assert out["data"].chunks == (1, 4, 4, 3)
            with h5py.File(alt_path, "r") as alt:
                assert np.all(alt["data"][...] == raw_data)
                assert alt["data"].chunks == (1, 4, 4, 3)
            # check hashes
            inp_hash = hash_file(inp_path, hash_type="md5", hash_mode="full", missing_ok=False)
            out_hash = hash_file(out_path, hash_type="md5", hash_mode="full", missing_ok=False)
            alt_hash = hash_file(alt_path, hash_type="md5", hash_mode="full", missing_ok=False)
            assert inp_hash == "8f21382c20cd0230aac5b1105fee8b39"
            assert out_hash == target_hash
            assert alt_hash == target_hash


def test_hdf5_speed_test():
    with create_temp_h5data(chunks=(_TEST_LEN, 4, 4, 3)) as (path, _):
        hdf5_test_speed(path, dataset_name="data", access_method="random")
    with create_temp_h5data(chunks=(1, 4, 4, 3)) as (path, _):
        hdf5_test_speed(path, dataset_name="data", access_method="sequential")
    with create_temp_h5data(chunks=None) as (path, _):
        hdf5_test_speed(path, dataset_name="data", access_method="sequential")


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import math

from disent.util.strings import colors as c

# ========================================================================= #
# Byte Formatting                                                           #
# ========================================================================= #


_BYTES_COLR = (c.WHT, c.lGRN, c.lYLW, c.lRED, c.lRED, c.lRED, c.lRED, c.lRED, c.lRED)
_BYTES_NAME = {
    1024: ("B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"),
    1000: ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"),
}


def bytes_to_human(size_bytes: int, decimals: int = 3, color: bool = True, mul: int = 1024) -> str:
    if size_bytes == 0:
        return "0B"
    if mul not in _BYTES_NAME:
        raise ValueError(f"invalid bytes multiplier: {repr(mul)} must be one of: {list(_BYTES_NAME.keys())}")
    # round correctly
    i = int(math.floor(math.log(size_bytes, mul)))
    s = round(size_bytes / math.pow(mul, i), decimals)
    # generate string
    name = f"{_BYTES_COLR[i]}{_BYTES_NAME[mul][i]}{c.RST}" if color else f"{_BYTES_NAME[mul][i]}"
    # format string
    return f"{s:{4+decimals}.{decimals}f} {name}"


# ========================================================================= #
# STRINGS                                                                   #
# ========================================================================= #


def make_separator_str(text, header=None, width=100, char_v="#", char_h="=", char_corners=None):
    """
    function wraps text between two lines or inside a box with lines on either side.
    FROM: my obstacle_tower project
    """
    if char_corners is None:
        char_corners = char_v
    assert len(char_v) == len(char_corners)
    assert len(char_h) == 1
    import pprint
    import textwrap

    def append_wrapped(text):
        for line in text.splitlines():
            for wrapped in textwrap.wrap(line, w, tabsize=4) if line.strip() else [""]:
                lines.append(f"{char_v} {wrapped:{w}s} {char_v}")

    w = width - 4
    lines = []
    sep = f"{char_corners} {char_h*w} {char_corners}"
    lines.append(f"\n{sep}")
    if header:
        append_wrapped(header)
        lines.append(sep)
    if type(text) != str:
        text = pprint.pformat(text, width=w)
    append_wrapped(text)
    lines.append(f"{sep}\n")
    return "\n".join(lines)


def make_box_str(text, header=None, width=100, char_v="|", char_h="-", char_corners="#"):
    """
    like print_separator but is isntead a box
    FROM: my obstacle_tower project
    """
    return make_separator_str(text, header=header, width=width, char_v=char_v, char_h=char_h, char_corners=char_corners)


def concat_lines(*strings, sep=" | "):
    """
    Join multi-line strings together horizontally, with the
    specified separator between them.
    """

    def pad_width(lines):
        max_len = max(len(line) for line in lines)
        return [f"{s:{max_len}}" for s in lines]

    def pad_height(list_of_lines):
        max_lines = max(len(lines) for lines in list_of_lines)
        return [(lines + ([""] * (max_lines - len(lines)))) for lines in list_of_lines]

    list_of_lines = [str(string).splitlines() for string in strings]
    list_of_lines = pad_height(list_of_lines)
    list_of_lines = [pad_width(lines) for lines in list_of_lines]
    return "\n".join(sep.join(rows) for rows in zip(*list_of_lines))


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
# https://github.com/google-research/disentanglement_lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGES:
# - converted from tensorflow to pytorch
# - removed gin config
# - uses disent objects and classes
# - renamed functions

"""
Unsupervised scores based on code covariance and mutual information.
"""

import logging

import numpy as np
import scipy

from disent.dataset import DisentDataset
from disent.metrics import utils
from disent.metrics.utils import make_metric

log = logging.getLogger(__name__)


# ========================================================================= #
# Unsupervised Scores                                                       #
# ========================================================================= #


@make_metric("unsupervised", fast_kwargs=dict(num_train=2000))
def metric_unsupervised(dataset: DisentDataset, representation_function, num_train=10000, batch_size=16):
    """Computes unsupervised scores based on covariance and mutual information.
    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observations as input and
        outputs a dim_representation sized representation for each observation.
      random_state: Numpy random state used for randomness.
      artifact_dir: Optional path to directory where artifacts can be saved.
      num_train: Number of points used for training.
      batch_size: Batch size for sampling.
    Returns:
      Dictionary with scores.
    """
    log.debug("Generating training set.")
    mus_train, _ = utils.generate_batch_factor_code(dataset, representation_function, num_train, batch_size)
    num_codes = mus_train.shape[0]
    cov_mus = np.cov(mus_train)
    assert num_codes == cov_mus.shape[0]

    # Gaussian total correlation.
    gaussian_total_correlation = _gaussian_total_correlation(cov_mus)

    # Gaussian Wasserstein correlation.
    gaussian_wasserstein_correlation = _gaussian_wasserstein_correlation(cov_mus)
    gaussian_wasserstein_correlation_norm = gaussian_wasserstein_correlation / np.sum(np.diag(cov_mus))

    # Compute average mutual information between different factors.
    mus_discrete = utils.histogram_discretize(mus_train, num_bins=20)
    mutual_info_matrix = utils.discrete_mutual_info(mus_discrete, mus_discrete)
    np.fill_diagonal(mutual_info_matrix, 0)
    mutual_info_score = np.sum(mutual_info_matrix) / (num_codes**2 - num_codes)

    return {
        "unsup.mi_score": mutual_info_score,
        "unsup.gauss_total_corr": gaussian_total_correlation,
        "unsup.gauss_wasser_corr": gaussian_wasserstein_correlation,
        "unsup.gauss_wasser_corr_norm": gaussian_wasserstein_correlation_norm,
    }


def _gaussian_total_correlation(cov):
    """Computes the total correlation of a Gaussian with covariance matrix cov.
    We use that the total correlation is the KL divergence between the Gaussian
    and the product of its marginals. By design, the means of these two Gaussians
    are zero and the covariance matrix of the second Gaussian is equal to the
    covariance matrix of the first Gaussian with off-diagonal entries set to zero.
    Args:
      cov: Numpy array with covariance matrix.
    Returns:
      Scalar with total correlation.
    """
    return 0.5 * (np.sum(np.log(np.diag(cov))) - np.linalg.slogdet(cov)[1])


def _gaussian_wasserstein_correlation(cov):
    """Wasserstein L2 distance between Gaussian and the product of its marginals.
    Args:
      cov: Numpy array with covariance matrix.
    Returns:
      Scalar with score.
    """
    sqrtm = scipy.linalg.sqrtm(cov * np.expand_dims(np.diag(cov), axis=1))
    return 2 * np.trace(cov) - 2 * np.trace(sqrtm)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
import torch

# ========================================================================= #
# Reduction Strategies                                                      #
# ========================================================================= #


def loss_reduction_mean(x: torch.Tensor) -> torch.Tensor:
    return x.mean()


def loss_reduction_mean_sum(x: torch.Tensor) -> torch.Tensor:
    return torch.flatten(x, start_dim=1).sum(dim=-1).mean()


_LOSS_REDUCTION_STRATEGIES = {
    "mean": loss_reduction_mean,
    "mean_sum": loss_reduction_mean_sum,
}


def loss_reduction(tensor: torch.Tensor, reduction="mean"):
    return _LOSS_REDUCTION_STRATEGIES[reduction](tensor)


# ========================================================================= #
# Reduction Strategies                                                      #
# ========================================================================= #


def get_mean_loss_scale(x: torch.Tensor, reduction: str):
    # check the dimensions if given
    assert 2 <= x.ndim <= 4, "unsupported number of dims, must be one of: BxC, BxHxW, BxCxHxW"

    # get the loss scaling
    if reduction == "sum":
        return np.prod(x.shape[1:])  # MEAN(B, SUM(C x H x W))
    elif reduction == "mean":
        return 1
    else:
        raise KeyError("unsupported loss reduction mode")


# ========================================================================= #
# loss batch reduction                                                      #
# ========================================================================= #


_REDUCTION_FNS = {
    "mean": torch.mean,
    "sum": torch.sum,
}


# TODO: this is duplicated in research ... pairwise_loss
# applying this function and then taking the
# mean should give the same result as loss_reduction
def batch_loss_reduction(tensor: torch.Tensor, reduction_dtype=None, reduction="mean") -> torch.Tensor:
    # mean over final dims
    if tensor.ndim >= 2:
        tensor = torch.flatten(tensor, start_dim=1)  # (B, -1)
        tensor = _REDUCTION_FNS[reduction](tensor, dim=-1, dtype=reduction_dtype)
    # check result
    assert tensor.ndim == 1
    # done
    return tensor


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
import torch

# ========================================================================= #
# convolve2d                                                                #
# ========================================================================= #


def _check_conv2d_inputs(signal, kernel):
    assert signal.ndim == 4, f"signal has {repr(signal.ndim)} dimensions, must have 4 dimensions instead: BxCxHxW"
    assert (
        kernel.ndim == 2 or kernel.ndim == 4
    ), f"kernel has {repr(kernel.ndim)} dimensions, must have 2 or 4 dimensions instead: HxW or BxCxHxW"
    # increase kernel size
    if kernel.ndim == 2:
        kernel = kernel[None, None, ...]
    # check kernel is an odd size
    kh, kw = kernel.shape[-2:]
    assert kh % 2 != 0 and kw % 2 != 0, f"kernel dimension sizes must be odd: ({kh}, {kw})"
    # check that broadcasting does not adjust the signal shape... TODO: relax this limitation?
    assert torch.broadcast_shapes(signal.shape[:2], kernel.shape[:2]) == signal.shape[:2]
    # done!
    return signal, kernel


def torch_conv2d_channel_wise(signal, kernel):
    """
    Apply the kernel to each channel separately!
    """
    signal, kernel = _check_conv2d_inputs(signal, kernel)
    # split channels into singel images
    fsignal = signal.reshape(-1, 1, *signal.shape[2:])
    # convolve each channel image
    out = torch.nn.functional.conv2d(fsignal, kernel, padding=(kernel.size(-2) // 2, kernel.size(-1) // 2))
    # reshape into original
    return out.reshape(-1, signal.shape[1], *out.shape[2:])


def torch_conv2d_channel_wise_fft(signal, kernel):
    """
    The same as torch_conv2d_channel_wise, but apply the kernel using fft.
    This is much more efficient for large filter sizes.

    Reference implementation is from: https://github.com/pyro-ppl/pyro/blob/ae55140acfdc6d4eade08b434195234e5ae8c261/pyro/ops/tensor_utils.py#L187
    """
    signal, kernel = _check_conv2d_inputs(signal, kernel)
    # get last dimension sizes
    sig_shape = np.array(signal.shape[-2:])
    ker_shape = np.array(kernel.shape[-2:])
    # compute padding
    padded_shape = sig_shape + ker_shape - 1
    # Compute convolution using fft.
    f_signal = torch.fft.rfft2(signal, s=tuple(padded_shape))
    f_kernel = torch.fft.rfft2(kernel, s=tuple(padded_shape))
    result = torch.fft.irfft2(f_signal * f_kernel, s=tuple(padded_shape))
    # crop final result
    s = (padded_shape - sig_shape) // 2
    f = s + sig_shape
    crop = result[..., s[0] : f[0], s[1] : f[1]]
    # done...
    return crop


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
# https://github.com/google-research/disentanglement_lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGES:
# - converted from tensorflow to pytorch
# - removed gin config
# - uses disent objects and classes
# - renamed functions

"""
Implementation of the SAP score.
Based on "Variational Inference of Disentangled Latent Concepts from Unlabeled
Observations" (https://openreview.net/forum?id=H1kG7GZAW), Section 3.
"""

import logging

import numpy as np
from sklearn import svm

from disent.dataset import DisentDataset
from disent.metrics import utils
from disent.metrics.utils import make_metric

log = logging.getLogger(__name__)


# ========================================================================= #
# SAP Scores                                                                #
# ========================================================================= #


@make_metric("sap", fast_kwargs=dict(num_train=2000, num_test=1000))
def metric_sap(
    dataset: DisentDataset,
    representation_function,
    num_train=10000,
    num_test=5000,
    batch_size=16,
    continuous_factors=False,
):
    """Computes the SAP score.
    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observations as input and
        outputs a dim_representation sized representation for each observation.
      num_train: Number of points used for training.
      num_test: Number of points used for testing discrete variables.
      batch_size: Batch size for sampling.
      continuous_factors: Factors are continuous variable (True) or not (False).
    Returns:
      Dictionary with SAP score.
    """
    log.debug("Generating training set.")
    mus, ys = utils.generate_batch_factor_code(dataset, representation_function, num_train, batch_size)
    mus_test, ys_test = utils.generate_batch_factor_code(dataset, representation_function, num_test, batch_size)
    log.debug("Computing score matrix.")
    return _compute_sap(mus, ys, mus_test, ys_test, continuous_factors)


def _compute_sap(mus, ys, mus_test, ys_test, continuous_factors):
    """Computes score based on both training and testing codes and factors."""
    score_matrix = _compute_score_matrix(mus, ys, mus_test, ys_test, continuous_factors)
    # Score matrix should have shape [num_latents, num_factors].
    assert score_matrix.shape[0] == mus.shape[0]
    assert score_matrix.shape[1] == ys.shape[0]

    sap_score = _compute_avg_diff_top_two(score_matrix)
    log.debug("SAP score: %.2g", sap_score)
    return {"sap.score": sap_score}  # "compactness: SAP" -- Measuring Disentanglement: A Review of Metrics


def _compute_score_matrix(mus, ys, mus_test, ys_test, continuous_factors):
    """Compute score matrix as described in Section 3."""
    num_latents = mus.shape[0]
    num_factors = ys.shape[0]
    score_matrix = np.zeros([num_latents, num_factors])
    for i in range(num_latents):
        for j in range(num_factors):
            mu_i = mus[i, :]
            y_j = ys[j, :]
            if continuous_factors:
                # Attribute is considered continuous.
                cov_mu_i_y_j = np.cov(mu_i, y_j, ddof=1)
                cov_mu_y = cov_mu_i_y_j[0, 1] ** 2
                var_mu = cov_mu_i_y_j[0, 0]
                var_y = cov_mu_i_y_j[1, 1]
                if var_mu > 1e-12:
                    score_matrix[i, j] = cov_mu_y * 1.0 / (var_mu * var_y)
                else:
                    score_matrix[i, j] = 0.0
            else:
                # Attribute is considered discrete.
                mu_i_test = mus_test[i, :]
                y_j_test = ys_test[j, :]
                classifier = svm.LinearSVC(C=0.01, class_weight="balanced")
                classifier.fit(mu_i[:, np.newaxis], y_j)
                pred = classifier.predict(mu_i_test[:, np.newaxis])
                score_matrix[i, j] = np.mean(pred == y_j_test)
    return score_matrix


def _compute_avg_diff_top_two(matrix):
    sorted_matrix = np.sort(matrix, axis=0)
    return np.mean(sorted_matrix[-1, :] - sorted_matrix[-2, :])



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import NoReturn
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy as np

from disent.dataset.data import GroundTruthData
from disent.dataset.sampling import BaseDisentSampler
from disent.dataset.util.state_space import StateSpace
from disent.util.jit import try_njit

# ========================================================================= #
# Pretend We Are Walking Ground-Truth Factors Randomly                      #
# ========================================================================= #


class GroundTruthRandomWalkSampler(BaseDisentSampler):
    def uninit_copy(self) -> "GroundTruthRandomWalkSampler":
        return GroundTruthRandomWalkSampler(
            num_samples=self._num_samples,
            p_dist_max=self._p_dist_max,
            n_dist_max=self._n_dist_max,
        )

    def __init__(
        self,
        num_samples: int = 3,
        p_dist_max: int = 8,
        n_dist_max: int = 32,
    ):
        super().__init__(num_samples=num_samples)
        # checks
        assert num_samples in {1, 2, 3}, f"num_samples ({repr(num_samples)}) must be 1, 2 or 3"
        # save hparams
        self._num_samples = num_samples
        self._p_dist_max = p_dist_max
        self._n_dist_max = n_dist_max
        # dataset variable
        self._state_space: Optional[StateSpace] = None

    def _init(self, dataset: GroundTruthData):
        assert isinstance(
            dataset, GroundTruthData
        ), f"dataset must be an instance of {repr(GroundTruthData.__class__.__name__)}, got: {repr(dataset)}"
        self._state_space = dataset.state_space_copy()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Sampling                                                              #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _sample_idx(self, idx) -> Tuple[int, ...]:
        if self._num_samples == 1:
            return (idx,)
        elif self._num_samples == 2:
            p_dist = np.random.randint(1, self._p_dist_max + 1)
            pos = _random_walk(idx, p_dist, self._state_space.factor_sizes)
            return (idx, pos)
        elif self._num_samples == 3:
            p_dist = np.random.randint(1, self._p_dist_max + 1)
            n_dist = np.random.randint(1, self._n_dist_max + 1)
            pos = _random_walk(idx, p_dist, self._state_space.factor_sizes)
            neg = _random_walk(pos, n_dist, self._state_space.factor_sizes)
            return (idx, pos, neg)
        else:
            raise RuntimeError


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _random_walk(idx: int, dist: int, factor_sizes: np.ndarray) -> int:
    # random walk
    pos = np.array(
        np.unravel_index(idx, factor_sizes), dtype=int
    )  # much faster than StateSpace.idx_to_pos, we don't need checks!
    for _ in range(dist):
        _walk_nearby_inplace(pos, factor_sizes)
    idx = np.ravel_multi_index(pos, factor_sizes)  # much faster than StateSpace.pos_to_idx, we don't need checks!
    # done!
    return int(idx)


@try_njit()
def _walk_nearby_inplace(pos: np.ndarray, factor_sizes: Sequence[int]) -> NoReturn:
    # try to shift any single factor by 1 or -1
    while True:
        f_idx = np.random.randint(0, len(factor_sizes))
        cur = pos[f_idx]
        # walk random factor value
        if np.random.random() < 0.5:
            nxt = max(cur - 1, 0)
        else:
            nxt = min(cur + 1, factor_sizes[f_idx] - 1)
        # exit if different
        if cur != nxt:
            break
    # update the position
    pos[f_idx] = nxt


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from typing import Sequence

import numpy as np
import torch
from torch import Tensor
from torch.distributions import Normal

from disent.frameworks.vae._unsupervised__vae import Vae

# ========================================================================= #
# InfoVae                                                                   #
# ========================================================================= #


class InfoVae(Vae):
    """
    InfoVAE: Balancing Learning and Inference in Variational Autoencoders
    https://arxiv.org/pdf/1706.02262.pdf

    TODO: this is not verified
    Reference implementation is from: https://github.com/AntixK/PyTorch-VAE
    Changes:
        1. kernels are computed weirdly in this implementation
        2. uses unbiased MMD estimates from https://arxiv.org/pdf/1505.03906.pdf
        3. computes means, not sums
    """

    REQUIRED_OBS = 1

    @dataclass
    class cfg(Vae.cfg):
        info_alpha: float = -0.5
        info_lambda: float = 5.0
        info_kernel: str = "rbf"
        # what is this? I don't think this should be configurable
        z_var: float = 2.0
        # this is optional
        maintain_reg_ratio: bool = True

    def __init__(self, model: "AutoEncoder", cfg: cfg = None, batch_augment=None):
        super().__init__(model=model, cfg=cfg, batch_augment=batch_augment)
        # checks
        assert self.cfg.info_alpha <= 0, f"cfg.info_alpha must be <= zero, current value is: {self.cfg.info_alpha}"
        assert self.cfg.loss_reduction == "mean", 'InfoVAE only supports cfg.loss_reduction == "mean"'

    # --------------------------------------------------------------------- #
    # Overrides                                                             #
    # --------------------------------------------------------------------- #

    def compute_ave_reg_loss(self, ds_posterior: Sequence[Normal], ds_prior: Sequence[Normal], zs_sampled):
        """
        TODO: This could be wrong?
        """

        # only supports one input observation at the moment
        (d_posterior,), (d_prior,), (z_sampled,) = ds_posterior, ds_prior, zs_sampled

        # compute kl divergence
        # compute maximum-mean discrepancy
        kl_loss = self.latents_handler.compute_ave_kl_loss(ds_posterior, ds_prior, zs_sampled)
        mmd_loss = self._compute_mmd(z_posterior_samples=z_sampled, z_prior_samples=d_prior.rsample())

        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
        # original loss Sums Everything, we use the mean and scale everything to keep the ratios the same
        # OLD: (C*W*H) * recon_mean + (Z)         * kl_mean + (Z)         * mmd_mean
        # NEW:           recon_mean + (Z)/(C*W*H) * kl_mean + (Z)/(C*W*H) * mmd_mean
        # compute the weight
        # TODO: maybe this should be standardised to something like Z=9, W=64, H=64, C=3
        # TODO: this could be moved into other models
        reg_weight = (self._model.z_size / np.prod(self._model.x_shape)) if self.cfg.maintain_reg_ratio else 1.0
        # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

        # weight the loss terms
        kl_reg_loss = reg_weight * (1 - self.cfg.info_alpha) * kl_loss
        mmd_reg_loss = reg_weight * (self.cfg.info_alpha + self.cfg.info_lambda - 1) * mmd_loss
        # compute combined loss
        combined_loss = kl_reg_loss + mmd_reg_loss

        # return logs
        return combined_loss, {
            "kl_loss": kl_loss,
            "kl_reg_loss": kl_reg_loss,
            "mmd_loss": mmd_loss,
            "mmd_reg_loss": mmd_reg_loss,
        }

    def _compute_mmd(self, z_posterior_samples: Tensor, z_prior_samples: Tensor) -> Tensor:
        """
        (✓) visual inspection against:
            https://en.wikipedia.org/wiki/Kernel_embedding_of_distributions#Kernel_two-sample_test
        """
        # check sizes - these conditions can be relaxed in practice, just for debugging
        assert z_posterior_samples.ndim == 2
        assert z_posterior_samples.shape == z_prior_samples.shape
        # compute kernels: (B, Z) -> (,)
        mean_pz_pz = self._compute_unbiased_mean(self._compute_kernel(z_prior_samples, z_prior_samples), unbaised=True)
        mean_pz_qz = self._compute_unbiased_mean(
            self._compute_kernel(z_prior_samples, z_posterior_samples), unbaised=False
        )
        mean_qz_qz = self._compute_unbiased_mean(
            self._compute_kernel(z_posterior_samples, z_posterior_samples), unbaised=True
        )
        # maximum-mean discrepancy
        mmd = mean_pz_pz - 2 * mean_pz_qz + mean_qz_qz
        return mmd

    def _compute_unbiased_mean(self, kernel: Tensor, unbaised: bool) -> Tensor:
        """
        (✓) visual inspection against equation (8) of
            Training generative neural networks via Maximum Mean Discrepancy optimization
            https://arxiv.org/pdf/1505.03906.pdf
        """
        # (B, B) == (N, M) ie. N=B and M=B
        N, M = kernel.shape
        assert N == M
        # compute mean along first and second dims
        if unbaised:
            # diagonal stacks values along last dimension ie. (B, B, Z) -> (Z, B) or (B, B) -> (B,)
            sum_kernel = kernel.sum(dim=(0, 1)) - torch.diagonal(kernel, dim1=0, dim2=1).sum(dim=-1)  # (B, B,) -> (,)
            # compute unbiased mean
            mean_kernel = sum_kernel / (N * (N - 1))
        else:
            mean_kernel = kernel.mean(dim=(0, 1))  # (B, B,) -> (,)
        # check size again
        assert mean_kernel.ndim == 0
        return mean_kernel

    def _compute_kernel(self, z0: Tensor, z1: Tensor) -> Tensor:
        """
        (✓) visual inspection against:
            https://en.wikipedia.org/wiki/Kernel_embedding_of_distributions#Kernel_two-sample_test
        """
        batch_size, z_size = z0.shape
        # convert tensors
        z0 = z0.unsqueeze(-2)  # convert to column tensor  # [B, Z] -> [B, 1, Z]
        z1 = z1.unsqueeze(-3)  # convert to row tensor     # [B, Z] -> [1, B, Z]
        # in our case this is not required, however it is useful
        # if z0 and z1 have different sizes along the 0th dimension.
        z0 = z0.expand(batch_size, batch_size, z_size)  # [B, 1, Z] -> [B, B, Z]
        z1 = z1.expand(batch_size, batch_size, z_size)  # [1, B, Z] -> [B, B, Z]
        # compute correct kernel
        if self.cfg.info_kernel == "rbf":
            kernel = self._kernel_rbf(z0, z1)
        # elif self.cfg.info_kernel == 'imq':
        #     kernel = self._kernel_imq(z0, z1)
        else:  # pragma: no cover
            raise KeyError(f"invalid cfg.info_kernel: {self.cfg.info_kernel}")
        # check result size
        assert kernel.shape == (batch_size, batch_size)
        return kernel

    def _kernel_rbf(self, x: Tensor, y: Tensor) -> Tensor:
        """
        Radial Basis Function (RBF) Kernel a.k.a. Gaussian Kernel
        k(x, y) = exp(- ||x - y||^2 / (2*sigma^2))

        (✓) visual inspection against:
            https://en.wikipedia.org/wiki/Reproducing_kernel_Hilbert_space#Radial_basis_function_kernels

        TODO: how do we arrive at the value for sigma?
              - multiplying sigma by z_size is that same as computing .mean(dim=-1)
                instead of the current sum
        TODO: do we treat each latent variable separately? or as vectors like now due to the .sum?
        """
        z_size = x.shape[-1]
        sigma = 2 * self.cfg.z_var * z_size
        kernel = torch.exp(-((x - y).pow(2).sum(dim=-1) / sigma))
        return kernel

    # def _kernel_imq(self, x: Tensor, y: Tensor, eps: float = 1e-7) -> Tensor:
    #     """
    #     Inverse Multi-Quadratics Kernel
    #     k(x, y) = (c^2 + ||x - y||^2)^b
    #         c ∈ R
    #         b < 0 but better if b ∈ (0, 1)
    #
    #     TODO: This could be wrong?
    #     # TODO: how do we arrive at the value for c
    #     """
    #     z_size = x.shape[-1]
    #     c = 2 * self.cfg.z_var * z_size
    #     kernel = c / (eps + c + (x - y).pow(2).sum(-1))
    #     return kernel


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import time
from contextlib import ContextDecorator
from math import log10

log = logging.getLogger(__name__)


# ========================================================================= #
# Memory Usage                                                              #
# ========================================================================= #


def get_memory_usage(pretty: bool = False):
    import os

    import psutil

    process = psutil.Process(os.getpid())
    num_bytes = process.memory_info().rss  # in bytes
    # format the bytes
    if pretty:
        from disent.util.strings.fmt import bytes_to_human

        return bytes_to_human(num_bytes)
    else:
        return num_bytes


# ========================================================================= #
# Context Manager Timer                                                     #
# ========================================================================= #


class Timer(ContextDecorator):

    """
    Timer class, can be used with a with statement to
    measure the execution time of a block of code!

    Examples:

        1. get the runtime
        ```
        with Timer() as t:
            time.sleep(1)
        print(t.pretty)
        ```

        2. automatic print
        ```
        with Timer(name="example") as t:
            time.sleep(1)
        ```

        3. reuse timer to measure multiple instances
        ```
        t = Timer()
        for i in range(100):
            with t:
                time.sleep(0.95)
            if t.elapsed > 3:
                break
        print(t)
        ```
    """

    def __init__(self, name: str = None, log_level: int = logging.INFO):
        self._start_time: int = None
        self._end_time: int = None
        self._total_time = 0
        self.name = name
        self._log_level = log_level

    def __enter__(self):
        self._start_time = time.time_ns()
        return self

    def __exit__(self, *args, **kwargs):
        self._end_time = time.time_ns()
        # add elapsed time to total time, and reset the timer!
        self._total_time += self._end_time - self._start_time
        self._start_time = None
        self._end_time = None
        # print results
        if self.name:
            if self._log_level is None:
                print(f"{self.name}: {self.pretty}")
            else:
                log.log(self._log_level, f"{self.name}: {self.pretty}")

    def restart(self):
        assert self._start_time is not None, "timer must have been started before we can restart it"
        assert self._end_time is None, "timer cannot be restarted if it is finished"
        self._start_time = time.time_ns()

    @property
    def elapsed_ns(self) -> int:
        if self._start_time is not None:
            # running
            return self._total_time + (time.time_ns() - self._start_time)
        # finished running
        return self._total_time

    @property
    def elapsed_ms(self) -> float:
        return self.elapsed_ns / 1_000_000

    @property
    def elapsed(self) -> float:
        return self.elapsed_ns / 1_000_000_000

    @property
    def pretty(self) -> str:
        return Timer.prettify_time(self.elapsed_ns)

    def __int__(self):
        return self.elapsed_ns

    def __float__(self):
        return self.elapsed

    def __str__(self):
        return self.pretty

    def __repr__(self):
        return self.pretty

    @staticmethod
    def prettify_time(ns: int) -> str:
        if ns == 0:
            return "N/A"
        elif ns < 0:
            return "NaN"
        # get power of 1000
        pow = min(3, int(log10(ns) // 3))
        time = ns / 1000**pow
        # get pretty string!
        if pow < 3 or time < 60:
            # less than 1 minute
            name = ["ns", "µs", "ms", "s"][pow]
            return f"{time:.3f}{name}"
        else:
            # 1 or more minutes
            s = int(time)
            d, s = divmod(s, 86400)
            h, s = divmod(s, 3600)
            m, s = divmod(s, 60)
            if d > 0:
                return f"{d}d:{h}h:{m}m"
            elif h > 0:
                return f"{h}h:{m}m:{s}s"
            else:
                return f"{m}m:{s}s"


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import contextlib
import os
import sys
from contextlib import contextmanager
from typing import Any
from typing import Dict

# ========================================================================= #
# TEST UTILS                                                                #
# ========================================================================= #


@contextmanager
def no_stdout():
    old_stdout = sys.stdout
    sys.stdout = open(os.devnull, "w")
    yield
    sys.stdout = old_stdout


@contextmanager
def no_stderr():
    old_stderr = sys.stderr
    sys.stderr = open(os.devnull, "w")
    yield
    sys.stderr = old_stderr


@contextlib.contextmanager
def temp_wd(new_wd):
    old_wd = os.getcwd()
    os.chdir(new_wd)
    yield
    os.chdir(old_wd)


@contextlib.contextmanager
def temp_sys_args(new_argv):
    # TODO: should this copy values?
    old_argv = sys.argv
    sys.argv = new_argv
    yield
    sys.argv = old_argv


@contextmanager
def temp_environ(environment: Dict[str, Any]):
    # TODO: should this copy values? -- could use unittest.mock.patch.dict(...)
    # save the old environment
    existing_env = {}
    for k in environment:
        if k in os.environ:
            existing_env[k] = os.environ[k]
    # update the environment
    os.environ.update(environment)
    # run the context
    try:
        yield
    finally:
        # restore the original environment
        for k in environment:
            if k in existing_env:
                os.environ[k] = existing_env[k]
            else:
                del os.environ[k]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import os
import re
import warnings
from numbers import Number
from typing import List
from typing import Tuple
from typing import Union

import numpy as np
import torch

import disent.registry as R
from disent.nn.functional import torch_box_kernel_2d
from disent.nn.functional import torch_conv2d_channel_wise_fft
from disent.nn.functional import torch_gaussian_kernel_2d
from disent.nn.modules import DisentModule

# ========================================================================= #
# Transforms                                                                #
# ========================================================================= #


TorLorN = Union[Number, Tuple[Number, Number], List[Number], np.ndarray]
MmTuple = Union[TorLorN, Tuple[TorLorN, TorLorN], List[TorLorN], np.ndarray]


def _expand_to_min_max_tuples(input: MmTuple) -> Tuple[Tuple[Number, Number], Tuple[Number, Number]]:
    (xm, xM), (ym, yM) = np.broadcast_to(input, (2, 2)).tolist()
    if not all(isinstance(n, (float, int)) for n in [xm, xM, ym, yM]):
        raise ValueError(
            "only scalars, tuples with shape (2,): [m, M] or tuples with shape (2, 2): [[xm, xM], [ym, yM]] are supported"
        )
    return (xm, xM), (ym, yM)


class _BaseFftBlur(DisentModule):
    """
    randomly gaussian blur the input images.
    - similar api to kornia
    """

    def __init__(self, p: float = 0.5, random_mode="batch", random_same_xy=True):
        super().__init__()
        # check arguments
        assert 0 <= p <= 1, f"probability of applying transform p={repr(p)} must be in range [0, 1]"
        self.p = p
        # random modes
        self.ran_batch, self.ran_channels = {
            "same": (False, False),
            "batch": (True, False),
            "all": (True, True),
            "channels": (False, True),
        }[random_mode]
        # same random value for x and y
        self.b_idx = 0 if random_same_xy else 1

    def forward(self, obs):
        # randomly return original
        if np.random.random() < (1 - self.p):
            return obs
        # add or remove batch dim
        add_batch_dim = obs.ndim == 3
        if add_batch_dim:
            obs = obs[None, ...]
        # apply kernel
        kernel = self._make_kernel(obs.shape, obs.device)
        result = torch_conv2d_channel_wise_fft(signal=obs, kernel=kernel)
        # remove batch dim
        if add_batch_dim:
            result = result[0]
        # done!
        return result

    def _make_kernel(self, shape, device):
        raise NotImplementedError


class FftGaussianBlur(_BaseFftBlur):
    """
    randomly gaussian blur the input images.
    - similar api to kornia
    """

    def __init__(
        self, sigma: MmTuple = 1.0, truncate: MmTuple = 3.0, p: float = 0.5, random_mode="batch", random_same_xy=True
    ):
        super().__init__(p=p, random_mode=random_mode, random_same_xy=random_same_xy)
        self.sigma = _expand_to_min_max_tuples(sigma)
        self.trunc = _expand_to_min_max_tuples(truncate)
        # same random value for x and y
        if random_same_xy:
            assert self.sigma[0] == self.sigma[1]
            assert self.trunc[0] == self.trunc[1]

    def _make_kernel(self, shape, device):
        B, C, H, W = shape
        # sigma & truncate
        sigma_m, sigma_M = torch.as_tensor(self.sigma, device=device).T
        trunc_m, trunc_M = torch.as_tensor(self.trunc, device=device).T
        # generate random values
        sigma = sigma_m + torch.rand(
            (B if self.ran_batch else 1), (C if self.ran_channels else 1), 2, dtype=torch.float32, device=device
        ) * (sigma_M - sigma_m)
        trunc = trunc_m + torch.rand(
            (B if self.ran_batch else 1), (C if self.ran_channels else 1), 2, dtype=torch.float32, device=device
        ) * (trunc_M - trunc_m)
        # generate kernel
        return torch_gaussian_kernel_2d(
            sigma=sigma[..., 0],
            truncate=trunc[..., 0],
            sigma_b=sigma[..., self.b_idx],
            truncate_b=trunc[..., self.b_idx],  # TODO: we do generate unneeded random values if random_same_xy == True
            dtype=torch.float32,
            device=device,
        )


class FftBoxBlur(_BaseFftBlur):
    """
    randomly box blur the input images.
    - similar api to kornia
    """

    def __init__(self, radius: MmTuple = 1, p: float = 0.5, random_mode="batch", random_same_xy=True):
        super().__init__(p=p, random_mode=random_mode, random_same_xy=random_same_xy)
        self.radius: Tuple[Tuple[int, int], Tuple[int, int]] = _expand_to_min_max_tuples(radius)
        # same random value for x and y
        if random_same_xy:
            assert self.radius[0] == self.radius[1]
        # check values
        values = np.array(self.radius).flatten().tolist()
        assert all(isinstance(x, int) for x in values), "radius values must be integers"
        assert all((0 <= x) for x in values), "radius values must be >= 0, resulting in diameter: 2*r+1"

    def _make_kernel(self, shape, device):
        B, C, H, W = shape
        # sigma & truncate
        (rym, ryM), (rxm, rxM) = self.radius
        # generate random values
        radius_y = torch.randint(
            low=rym, high=ryM + 1, size=((B if self.ran_batch else 1), (C if self.ran_channels else 1)), device=device
        )
        radius_x = torch.randint(
            low=rxm, high=rxM + 1, size=((B if self.ran_batch else 1), (C if self.ran_channels else 1)), device=device
        )
        # done computing kernel
        return torch_box_kernel_2d(
            radius=radius_y, radius_b=radius_x if (self.b_idx == 1) else radius_y, dtype=torch.float32, device=device
        )


# ========================================================================= #
# FFT Kernel                                                                #
# ========================================================================= #


_NO_ARG = object()


class FftKernel(DisentModule):
    """
    2D Convolve an image
    """

    def __init__(self, kernel: Union[torch.Tensor, str], normalize_mode: str = _NO_ARG):
        super().__init__()
        # deprecation error
        if normalize_mode is _NO_ARG:
            raise ValueError(
                f'default argument for normalize_mode was "sum", this has been deprecated and will change to "none" in future. Please manually override this value!'
            )
        # load & save the kernel -- no gradients allowed
        self._kernel: torch.Tensor
        self.register_buffer("_kernel", get_kernel(kernel, normalize_mode=normalize_mode), persistent=True)
        self._kernel.requires_grad = False

    def forward(self, obs):
        # add or remove batch dim
        add_batch_dim = obs.ndim == 3
        if add_batch_dim:
            obs = obs[None, ...]
        # apply kernel
        result = torch_conv2d_channel_wise_fft(signal=obs, kernel=self._kernel)
        # remove batch dim
        if add_batch_dim:
            result = result[0]
        # done!
        return result


# ========================================================================= #
# Kernels                                                                   #
# ========================================================================= #


@torch.no_grad()
def _scale_kernel(kernel: torch.Tensor, mode: Union[bool, str] = "abssum"):
    # old normalize mode
    if isinstance(mode, bool):
        raise ValueError(
            f'boolean arguments to `scale_kernel` are deprecated, convert True to "sum" and False to "none", got: {repr(mode)}'
        )
    # handle the normalize mode
    if mode == "sum":
        return kernel / kernel.sum()
    elif mode == "abssum":
        return kernel / torch.abs(kernel).sum()
    elif mode == "possum":
        return kernel / torch.abs(kernel)[kernel > 0].sum()
    elif mode == "negsum":
        return kernel / torch.abs(kernel)[kernel < 0].sum()
    elif mode == "maxsum":
        return kernel / torch.maximum(
            torch.abs(kernel)[kernel > 0].sum(),
            torch.abs(kernel)[kernel < 0].sum(),
        )
    elif mode == "none":
        return kernel
    else:
        raise KeyError(f"invalid scale mode: {repr(mode)}")


def _check_kernel(kernel: torch.Tensor) -> torch.Tensor:
    # check kernel
    assert isinstance(kernel, torch.Tensor)
    assert kernel.dtype == torch.float32
    assert kernel.ndim == 4, f"invalid number of kernel dims, required 4, given: {repr(kernel.ndim)}"  # B, C, H, W
    assert (
        kernel.shape[0] == 1
    ), f"invalid size of first kernel dim, required (1, ?, ?, ?), given: {repr(kernel.shape)}"  # B
    assert kernel.shape[0] in (
        1,
        3,
    ), f"invalid size of second kernel dim, required (?, 1 or 3, ?, ?), given: {repr(kernel.shape)}"  # C
    # done checks
    return kernel


# NOTE: this function compliments make_reconstruction_loss in frameworks/helper/reconstructions.py
def make_kernel(name: str, normalize_mode: str = "none"):
    kernel = R.KERNELS[name]
    kernel = _scale_kernel(kernel, mode=normalize_mode)
    kernel = _check_kernel(kernel)
    return kernel


def _get_kernel(name_or_path: str) -> torch.Tensor:
    if "/" not in name_or_path:
        try:
            return R.KERNELS[name_or_path]
        except KeyError:
            pass
    if os.path.isfile(name_or_path):
        return torch.load(name_or_path)
    raise KeyError(
        f"Invalid kernel path or name: {repr(name_or_path)} Examples of argument based kernels include: {R.KERNELS.regex_examples}, otherwise specify a valid path to a kernel file save with torch."
    )


def get_kernel(kernel: Union[str, torch.Tensor], normalize_mode: str = "none"):
    kernel = _get_kernel(kernel) if isinstance(kernel, str) else torch.clone(kernel)
    kernel = _scale_kernel(kernel, mode=normalize_mode)
    kernel = _check_kernel(kernel)
    return kernel


# ========================================================================= #
# Registered Kernels                                                        #
# ========================================================================= #


# we register this in disent.registry
def _make_box_kernel(radius: str):
    return torch_box_kernel_2d(radius=int(radius))[None, ...]


# we register this in disent.registry
def _make_gaussian_kernel(radius: str):
    return torch_gaussian_kernel_2d(sigma=int(radius) / 4.0, truncate=4.0)[None, None, ...]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
# https://github.com/google-research/disentanglement_lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGES:
# - converted from tensorflow to pytorch
# - removed gin config
# - uses disent objects and classes
# - renamed functions

"""
Mutual Information Gap from the beta-TC-VAE paper.
Based on "Isolating Sources of Disentanglement in Variational Autoencoders"
(https://arxiv.org/pdf/1802.04942.pdf).
"""

import logging

import numpy as np

from disent.dataset import DisentDataset
from disent.metrics import utils
from disent.metrics.utils import make_metric

log = logging.getLogger(__name__)


# ========================================================================= #
# Mutual Information Gap                                                    #
# ========================================================================= #


@make_metric("mig", fast_kwargs=dict(num_train=2000))
def metric_mig(
    dataset: DisentDataset,
    representation_function,
    num_train=10000,
    batch_size=16,
):
    """Computes the mutual information gap.
    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observations as input and
        outputs a dim_representation sized representation for each observation.
      num_train: Number of points used for training.
      batch_size: Batch size for sampling.
    Returns:
      Dict with average mutual information gap.
    """
    log.debug("Generating training set.")
    mus_train, ys_train = utils.generate_batch_factor_code(dataset, representation_function, num_train, batch_size)
    assert mus_train.shape[1] == num_train
    return _compute_mig(mus_train, ys_train)


def _compute_mig(mus_train, ys_train):
    """
    Computes score based on both training and testing codes and factors.
    """
    discretized_mus = utils.histogram_discretize(mus_train, num_bins=20)
    m = utils.discrete_mutual_info(discretized_mus, ys_train)
    assert m.shape[0] == mus_train.shape[0]
    assert m.shape[1] == ys_train.shape[0]
    # m is [num_latents, num_factors]
    entropy = utils.discrete_entropy(ys_train)
    sorted_m = np.sort(m, axis=0)[::-1]
    return {
        "mig.discrete_score": np.mean(
            np.divide(sorted_m[0, :] - sorted_m[1, :], entropy[:])
        )  # "modularity: MIG" -- Measuring Disentanglement: A Review of Metrics
    }



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

"""
Factored Components Metric
- Michlo et al.
  https://github.com/nmichlo/msc-research
"""

import logging
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import torch
import torch.nn.functional as F
from scipy.stats import pearsonr
from scipy.stats import spearmanr

from disent.dataset import DisentDataset
from disent.frameworks.helper.reconstructions import ReconLossHandler
from disent.metrics._flatness import encode_all_along_factor
from disent.metrics._flatness import encode_all_factors
from disent.metrics._flatness import filter_inactive_factors
from disent.metrics.utils import make_metric
from disent.nn.functional import torch_mean_generalized
from disent.nn.functional import torch_pca
from disent.nn.loss.reduction import batch_loss_reduction
from disent.util import to_numpy

log = logging.getLogger(__name__)


# ========================================================================= #
# flatness                                                                  #
# ========================================================================= #


_SAMPLES_MULTIPLIER_GLOBAL = 4
_SAMPLES_MULTIPLIER_FACTOR = 2


# ========================================================================= #
# flatness                                                                  #
# ========================================================================= #


def _metric_factored_components(
    dataset: DisentDataset,
    representation_function: callable,
    num_samples: int = 64,
    global_subset_size: int = 32,
    repeats: int = 1024,
    batch_size: int = 64,
    compute_distances: bool = True,
    compute_linearity: bool = True,
):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    Computes the factored components metric (ordering, linearity & axis alignment):

    # Distances:
        rcorr_factor_data:   rank correlation between ground-truth factor dists and MSE distances between data points
        rcorr_latent_data:   rank correlation between l2 latent dists           and MSE distances between data points
        rcorr_factor_latent: rank correlation between ground-truth factor dists and latent dists

        lcorr_factor_data:   linear correlation between ground-truth factor dists and MSE distances between data points
        lcorr_latent_data:   linear correlation between l2 latent dists           and MSE distances between data points
        lcorr_factor_latent: linear correlation between ground-truth factor dists and latent dists

        -- only active if `compute_swap_ratio=True`
        rsame_factor_data:   how similar ground-truth factor dists are compared to MSE distances between data points  MEAN: ((a<A)&(b<B)) | ((a==A)&(b==B)) | ((a>A)&(b>B))
        rsame_latent_data:   how similar l2 latent dists           are compared to MSE distances between data points  MEAN: ((a<A)&(b<B)) | ((a==A)&(b==B)) | ((a>A)&(b>B))
        rsame_factor_latent: how similar ground-truth factor dists are compared to latent dists                       MEAN: ((a<A)&(b<B)) | ((a==A)&(b==B)) | ((a>A)&(b>B))

        * modifiers:
            - .global | computed using random global samples
            - .factor | computed using random values along a ground-truth factor traversal
            - .l1     | computed using l1 distance
            - .l2     | computed using l2 distance -- (if an .l1 or .l2 tag is missing, then it is .l2 by default)

    # Linearity & Axis Alignment
        linear_ratio:           average (largest singular value over sum of singular values)
        axis_ratio:             average (largest std/variance over sum of std/variances)
        axis_alignment:         axis ratio is bounded by linear ratio - compute: axis / linear

        linear_ratio_ave:       [INVALID] average (largest singular value) over average (sum of singular values)
        axis_ratio_ave:         average (largest std/variance) over average (sum of std/variances)
        axis_alignment_ave:     [INVALID] axis ratio is bounded by linear ratio - compute: axis / linear

        * modifiers:
            - .var | computed using the variance
            - .std | computed using the standard deviation

    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observations as input and outputs a dim_representation sized representation for each observation.
      num_samples: How many random triplets are sampled from a single factor-traversal or global ranom mini-batch
      global_subset_size: Controls the size of the global random minibatch, for individual factors this is usually the size of the factor. Triplets are randomly sampled from this.
      repeats: how many times to repeat a traversal along each factors, these are then averaged together.
      batch_size: Batch size to process at any time while generating representations, should not effect metric results.
      compute_distances: If the distance components of the metric should be computed.
      compute_linearity: If the linearity components of the metric should be computed.
    Returns:
      Dictionary with metrics
    """
    # checks
    if not (compute_distances or compute_linearity):
        raise ValueError(
            f"Metric will not compute any values! At least one of: `compute_distances` or `compute_linearity` must be `True`"
        )

    # compute actual metric values
    factor_scores, global_scores = _compute_factored_metric_components(
        dataset,
        representation_function,
        num_samples=num_samples,
        global_subset_size=global_subset_size,
        repeats=repeats,
        batch_size=batch_size,
        compute_distances=compute_distances,
        compute_linearity=compute_linearity,
    )

    # convert values from torch
    scores = {
        **global_scores,
        **factor_scores,
    }

    # sorted
    return {k: scores[k] for k in sorted(scores.keys())}


# EXPORT: metric_factored_components
metric_factored_components = make_metric(
    "factored_components",
    default_kwargs=dict(compute_distances=True, compute_linearity=True),
    fast_kwargs=dict(compute_distances=True, compute_linearity=True, repeats=128),
)(_metric_factored_components)

# EXPORT: metric_distances
metric_distances = make_metric(
    "distances",
    default_kwargs=dict(compute_distances=True, compute_linearity=False),
    fast_kwargs=dict(compute_distances=True, compute_linearity=False, repeats=128),
)(_metric_factored_components)

# EXPORT: metric_linearity
metric_linearity = make_metric(
    "linearity",
    default_kwargs=dict(compute_distances=False, compute_linearity=True),
    fast_kwargs=dict(compute_distances=False, compute_linearity=True, repeats=128),
)(_metric_factored_components)


# ========================================================================= #
# flatness                                                                  #
# ========================================================================= #


def _filtered_mean(values: torch.Tensor, p: Union[str, int], factor_sizes: Tuple[int, ...]):
    # increase precision
    values = values.to(torch.float64)
    # check size
    assert values.shape == (len(factor_sizes),)
    # filter
    # -- filter out factor dimensions that are incorrect. ie. size <= 1
    values = filter_inactive_factors(values, factor_sizes)
    # compute mean
    mean = torch_mean_generalized(values, dim=0, p=p)
    # return decreased precision
    return to_numpy(mean.to(torch.float32))


@torch.no_grad()
def _compute_factored_metric_components(
    dataset: DisentDataset,
    representation_function,
    num_samples: int,
    global_subset_size: int,
    repeats: int,
    batch_size: int,
    compute_distances: bool,
    compute_linearity: bool,
) -> (dict, dict):
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    # COMPUTE FOR EACH FACTOR
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

    # shapes: (num_factors,)
    factor_values: Dict[str, np.ndarray] = _numpy_stack_all_dicts(
        [
            _compute_factored_metric_components_along_factor(
                dataset,
                representation_function,
                f_idx=f_idx,
                num_samples=num_samples,
                repeats=repeats,
                batch_size=batch_size,
                compute_distances=compute_distances,
                compute_linearity=compute_linearity,
            )
            for f_idx in range(dataset.gt_data.num_factors)
        ]
    )

    # aggregate for each factor
    # -- filter out factor dimensions that are incorrect. ie. size <= 1
    factor_scores = {
        k: float(_filtered_mean(torch.from_numpy(v), p="geometric", factor_sizes=dataset.gt_data.factor_sizes))
        for k, v in factor_values.items()
    }

    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    # RANDOM GLOBAL SAMPLES
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

    if compute_distances:
        # storage
        distance_measures: List[Dict[str, np.ndarray]] = []

        # was: `iter_chunks(range(int(repeats * np.mean(dataset.gt_data.factor_sizes))), batch_size)`
        for _ in range(repeats):
            # sample random factors
            factors = dataset.gt_data.sample_factors(size=global_subset_size)
            # encode factors
            zs, xs = encode_all_factors(
                dataset, representation_function, factors, batch_size=batch_size, return_batch=True
            )
            zs, xs, factors = zs.cpu(), xs.cpu(), torch.from_numpy(factors).to(torch.float32)
            # [COMPUTE SAME RATIO & CORRELATION]: was `_SAMPLES_MULTIPLIER_GLOBAL*len(zs)`
            computed_dists = _compute_dists(num_triplets=num_samples, zs_traversal=zs, xs_traversal=xs, factors=factors)
            # [STORE DISTANCES]
            distance_measures.append(computed_dists)

        # [AGGREGATE]
        # concatenate all into arrays: <shape: (repeats*num_samples,)>
        # then aggregate over first dimension: <shape: (,)>
        distance_measures: Dict[str, np.ndarray] = _numpy_concat_all_dicts(distance_measures)
        distance_measures: Dict[str, float] = _compute_scores_from_dists(distance_measures)
        distance_measures: Dict[str, float] = {f"distances.{k}.global": v for k, v in distance_measures.items()}
    else:
        distance_measures: Dict[str, float] = {}

    # update global scores
    global_scores = distance_measures

    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    # RETURN
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

    return factor_scores, global_scores


# ========================================================================= #
# CORE                                                                      #
# -- using variance instead of standard deviation makes it easier to        #
#    obtain high scores.                                                    #
# ========================================================================= #


def _compute_unsorted_axis_values(zs_traversal, use_std: bool = True):
    # CORRELATIONS -- SORTED IN DESCENDING ORDER:
    # correlation with standard basis (1, 0, 0, ...), (0, 1, 0, ...), ...
    axis_values = torch.var(zs_traversal, dim=0)  # (z_size,)
    if use_std:
        axis_values = torch.sqrt(axis_values)
    return axis_values


def _compute_unsorted_linear_values(zs_traversal, use_std: bool = True):
    # CORRELATIONS -- SORTED IN DESCENDING ORDER:
    # correlation along arbitrary orthogonal basis
    # -- note pca_mode='svd' returns the number of values equal to: min(factor_size, z_size)  !!! this may lower scores on average
    # -- note pca_mode='eig' returns the number of values equal to: z_size
    _, linear_values = torch_pca(zs_traversal, center=True, mode="eig")
    if use_std:
        linear_values = torch.sqrt(linear_values)
    return linear_values


def _score_from_sorted(sorted_vars: torch.Tensor, top_2: bool = False, norm: bool = True) -> torch.Tensor:
    if top_2:
        # use two max values
        # this is more like mig
        sorted_vars = sorted_vars[:2]
    # sum all values
    n = len(sorted_vars)
    r = sorted_vars[0] / torch.sum(sorted_vars)
    # get norm if needed
    if norm:
        # for: x/(x+a)
        # normalised = (x/(x+a) - (1/n)) / (1 - (1/n))
        # normalised = (x - 1/(n-1) * a) / (x + a)
        r = (r - (1 / n)) / (1 - (1 / n))
    # done!
    return r


def _score_from_unsorted(unsorted_values: torch.Tensor, top_2: bool = False, norm: bool = True):
    assert unsorted_values.ndim == 1
    # sort in descending order
    sorted_values = torch.sort(unsorted_values, dim=-1, descending=True).values
    # compute score
    return _score_from_sorted(sorted_values, top_2=top_2, norm=norm)


def compute_axis_score(zs_traversal: torch.Tensor, use_std: bool = True, top_2: bool = False, norm: bool = True):
    unsorted_values = _compute_unsorted_axis_values(zs_traversal, use_std=use_std)
    score = _score_from_unsorted(unsorted_values, top_2=top_2, norm=norm)
    return score


def compute_linear_score(zs_traversal: torch.Tensor, use_std: bool = True, top_2: bool = False, norm: bool = True):
    unsorted_values = _compute_unsorted_linear_values(zs_traversal, use_std=use_std)
    score = _score_from_unsorted(unsorted_values, top_2=top_2, norm=norm)
    return score


# ========================================================================= #
# Collate Dict - Helper Functions - More specific than `default_collate()`  #
# ========================================================================= #


def _torch_concat_all_dicts(dists_list: List[Dict[str, torch.Tensor]]) -> Dict[str, torch.Tensor]:
    return {k: torch.cat([dists_dict[k] for dists_dict in dists_list], dim=0) for k in dists_list[0].keys()}


def _torch_stack_all_dicts(dists_list: List[Dict[str, torch.Tensor]]) -> Dict[str, torch.Tensor]:
    return {k: torch.stack([dists_dict[k] for dists_dict in dists_list], dim=0) for k in dists_list[0].keys()}


def _numpy_concat_all_dicts(dists_list: List[Dict[str, Union[np.ndarray, float, int]]]) -> Dict[str, np.ndarray]:
    return {k: np.concatenate([dists_dict[k] for dists_dict in dists_list], axis=0) for k in dists_list[0].keys()}


def _numpy_stack_all_dicts(dists_list: List[Dict[str, Union[np.ndarray, float, int]]]) -> Dict[str, np.ndarray]:
    return {k: np.stack([dists_dict[k] for dists_dict in dists_list], axis=0) for k in dists_list[0].keys()}


# ========================================================================= #
# Distance Helper Functions                                                 #
# ========================================================================= #


def _same_mask(ap0, an0, ap1, an1):
    return ((ap0 < an0) & (ap1 < an1)) | ((ap0 == an0) & (ap1 == an1)) | ((ap0 > an0) & (ap1 > an1))


def _unswapped_ratio_torch(ap0: torch.Tensor, an0: torch.Tensor, ap1: torch.Tensor, an1: torch.Tensor) -> float:
    # values must correspond
    same_mask = _same_mask(ap0=ap0, an0=an0, ap1=ap1, an1=an1)
    # num values
    return float(same_mask.to(torch.float32).mean())


def _unswapped_ratio_numpy(ap0: np.ndarray, an0: np.ndarray, ap1: np.ndarray, an1: np.ndarray):
    # values must correspond
    same_mask = _same_mask(ap0=ap0, an0=an0, ap1=ap1, an1=an1)
    # num values
    return np.mean(same_mask, dtype="float32")


def _compute_dists(
    num_triplets: int,
    zs_traversal: Optional[torch.Tensor],
    xs_traversal: torch.Tensor,
    factors: Optional[torch.Tensor],
    recon_loss_fn=F.mse_loss,
) -> Dict[str, np.ndarray]:
    assert (factors is None) or (len(factors) == len(xs_traversal))
    assert (zs_traversal is None) or (len(zs_traversal) == len(xs_traversal))

    # get the recon loss function
    def _unreduced_loss(input, target):
        if isinstance(recon_loss_fn, ReconLossHandler):
            return recon_loss_fn.compute_unreduced_loss(input, target)
        else:
            return recon_loss_fn(input, target, reduction="none")

    # compute!
    with torch.no_grad():
        # generate random triplets
        # - {p, n} indices do not need to be sorted like triplets, these can be random.
        #   This metric is symmetric for swapped p & n values.
        idxs_a, idxs_p, idxs_n = torch.randint(0, len(xs_traversal), size=(3, num_triplets), device=xs_traversal.device)
        # compute distances -- shape: (num,)
        distances = {
            "ap_ground_dists": (
                torch.norm(factors[idxs_a, :] - factors[idxs_p, :], p=1, dim=-1)
                if (factors is not None)
                else torch.abs(idxs_a - idxs_p)
            )
            .cpu()
            .numpy(),
            "an_ground_dists": (
                torch.norm(factors[idxs_a, :] - factors[idxs_n, :], p=1, dim=-1)
                if (factors is not None)
                else torch.abs(idxs_a - idxs_n)
            )
            .cpu()
            .numpy(),
            "ap_data_dists": batch_loss_reduction(
                _unreduced_loss(xs_traversal[idxs_a, ...], xs_traversal[idxs_p, ...]),
                reduction_dtype=torch.float32,
                reduction="mean",
            )
            .cpu()
            .numpy(),
            "an_data_dists": batch_loss_reduction(
                _unreduced_loss(xs_traversal[idxs_a, ...], xs_traversal[idxs_n, ...]),
                reduction_dtype=torch.float32,
                reduction="mean",
            )
            .cpu()
            .numpy(),
        }
        # compute distances -- shape: (num,)
        if zs_traversal is not None:
            distances.update(
                {
                    "ap_latent_dists.l1": torch.norm(zs_traversal[idxs_a, :] - zs_traversal[idxs_p, :], dim=-1, p=1)
                    .cpu()
                    .numpy(),
                    "an_latent_dists.l1": torch.norm(zs_traversal[idxs_a, :] - zs_traversal[idxs_n, :], dim=-1, p=1)
                    .cpu()
                    .numpy(),
                    "ap_latent_dists.l2": torch.norm(zs_traversal[idxs_a, :] - zs_traversal[idxs_p, :], dim=-1, p=2)
                    .cpu()
                    .numpy(),
                    "an_latent_dists.l2": torch.norm(zs_traversal[idxs_a, :] - zs_traversal[idxs_n, :], dim=-1, p=2)
                    .cpu()
                    .numpy(),
                }
            )
        # return values -- shape: (num,)
        return distances


def _compute_scores_from_dists(dists: Dict[str, np.array]) -> Dict[str, float]:
    # [DATA & GROUND DISTS]:
    # extract the distances -- shape: (num,)
    ap_ground_dists = dists["ap_ground_dists"]
    an_ground_dists = dists["an_ground_dists"]
    ap_data_dists = dists["ap_data_dists"]
    an_data_dists = dists["an_data_dists"]
    # concatenate values -- shape: (2 * num,)
    ground_dists = np.concatenate([ap_ground_dists, an_ground_dists], axis=0)
    data_dists = np.concatenate([ap_data_dists, an_data_dists], axis=0)
    # compute the scores
    # - check the number of swapped elements along a factor for random triplets.
    # - compute the spearman rank correlation coefficient over the concatenated distances
    # - compute the pearman correlation coefficient over the concatenated distances
    scores = {
        "rsame_ground_data": _unswapped_ratio_numpy(
            ap0=ap_ground_dists, an0=an_ground_dists, ap1=ap_data_dists, an1=an_data_dists
        ),  # simplifies to: (ap_data_dists > an_data_dists).to(torch.float32).mean()
        "rcorr_ground_data": spearmanr(ground_dists, data_dists)[0],
        "lcorr_ground_data": pearsonr(ground_dists, data_dists)[0],
    }

    # [RETURN EARLY]:
    if "ap_latent_dists.l1" not in dists:
        return scores

    # [LATENT DISTS]:
    # extract the distances -- shape: (num,)
    ap_latent_dists_l1 = dists["ap_latent_dists.l1"]
    an_latent_dists_l1 = dists["an_latent_dists.l1"]
    ap_latent_dists_l2 = dists["ap_latent_dists.l2"]
    an_latent_dists_l2 = dists["an_latent_dists.l2"]
    # concatenate values -- shape: (2 * num,)
    latent_dists_l1 = np.concatenate([ap_latent_dists_l1, an_latent_dists_l1], axis=0)
    latent_dists_l2 = np.concatenate([ap_latent_dists_l2, an_latent_dists_l2], axis=0)
    # compute the scores
    scores.update(
        {
            # - check the number of swapped elements along a factor for random triplets.
            "rsame_ground_latent.l1": _unswapped_ratio_numpy(
                ap0=ap_ground_dists, an0=an_ground_dists, ap1=ap_latent_dists_l1, an1=an_latent_dists_l1
            ),  # simplifies to: (ap_latent_dists > an_latent_dists).to(torch.float32).mean()
            "rsame_latent_data.l1": _unswapped_ratio_numpy(
                ap0=ap_latent_dists_l1, an0=an_latent_dists_l1, ap1=ap_data_dists, an1=an_data_dists
            ),
            "rsame_ground_latent.l2": _unswapped_ratio_numpy(
                ap0=ap_ground_dists, an0=an_ground_dists, ap1=ap_latent_dists_l2, an1=an_latent_dists_l2
            ),  # simplifies to: (ap_latent_dists > an_latent_dists).to(torch.float32).mean()
            "rsame_latent_data.l2": _unswapped_ratio_numpy(
                ap0=ap_latent_dists_l2, an0=an_latent_dists_l2, ap1=ap_data_dists, an1=an_data_dists
            ),
            # - compute the spearman rank correlation coefficient over the concatenated distances
            "rcorr_ground_latent.l1": spearmanr(ground_dists, latent_dists_l1)[0],
            "rcorr_latent_data.l1": spearmanr(latent_dists_l1, data_dists)[0],
            "rcorr_ground_latent.l2": spearmanr(ground_dists, latent_dists_l2)[0],
            "rcorr_latent_data.l2": spearmanr(latent_dists_l2, data_dists)[0],
            # - compute the pearman correlation coefficient over the concatenated distances
            "lcorr_ground_latent.l1": pearsonr(ground_dists, latent_dists_l1)[0],
            "lcorr_latent_data.l1": pearsonr(latent_dists_l1, data_dists)[0],
            "lcorr_ground_latent.l2": pearsonr(ground_dists, latent_dists_l2)[0],
            "lcorr_latent_data.l2": pearsonr(latent_dists_l2, data_dists)[0],
        }
    )

    # [DONE]
    return scores


# ========================================================================= #
# TRAVERSAL FLATNESS                                                        #
# ========================================================================= #


def _compute_factored_metric_components_along_factor(
    dataset: DisentDataset,
    representation_function,
    f_idx: int,
    num_samples: int,
    repeats: int,
    batch_size: int,
    compute_distances: bool,
    compute_linearity: bool,
) -> Dict[str, float]:
    # NOTE: what to do if the factor size is too small?

    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    # FEED FORWARD, COMPUTE ALL
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

    distance_measures: List[Dict[str, np.ndarray]] = []
    linear_measures: List[Dict[str, torch.Tensor]] = []

    for i in range(repeats):
        # [ENCODE TRAVERSAL]:
        # - generate repeated factors, varying one factor over the entire range
        # - shape: (factor_size, z_size)
        zs_traversal, xs_traversal = encode_all_along_factor(
            dataset, representation_function, f_idx=f_idx, batch_size=batch_size, return_batch=True
        )
        zs_traversal = zs_traversal.cpu()

        if compute_distances:
            xs_traversal = xs_traversal.cpu()
            # [COMPUTE SAME RATIO & CORRELATION] | was: `num_triplets=_SAMPLES_MULTIPLIER_FACTOR*len(zs_traversal)`
            computed_dists = _compute_dists(
                num_triplets=num_samples, zs_traversal=zs_traversal, xs_traversal=xs_traversal, factors=None
            )
            # [STORE DISTANCES]
            distance_measures.append(computed_dists)

        if compute_linearity:
            # [VARIANCE ALONG DIFFERING AXES]:
            # 1. axis: correlation with standard basis (1, 0, 0, ...), (0, 1, 0, ...), ...
            # 2. linear: correlation along arbitrary orthogonal bases
            axis_values_var = _compute_unsorted_axis_values(zs_traversal, use_std=False)  # shape: (z_size,)
            linear_values_var = _compute_unsorted_linear_values(zs_traversal, use_std=False)  # shape: (z_size,)
            # [COMPUTE LINEARITY SCORES]:
            axis_ratio_var = _score_from_unsorted(axis_values_var, top_2=False, norm=True)  # shape: ()
            linear_ratio_var = _score_from_unsorted(linear_values_var, top_2=False, norm=True)  # shape: ()
            # [STORE SCORES]
            linear_measures.append(
                {
                    "linearity.axis_ratio.var": axis_ratio_var,
                    "linearity.linear_ratio.var": linear_ratio_var,
                    # aggregating linear values outside this function does not make sense, values do not correspond between repeats.
                    "linearity.axis_alignment.var": axis_ratio_var / (linear_ratio_var + 1e-20),
                    # temp values
                    "_TEMP_.axis_values.var": axis_values_var,
                }
            )

    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    # AGGREGATE DATA - For each distance measure
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

    if compute_distances:
        # concatenate all into arrays: <shape: (repeats*num_samples,)>
        # then aggregate over first dimension: <shape: (,)>
        distance_measures: Dict[str, np.ndarray] = _numpy_concat_all_dicts(distance_measures)
        distance_measures: Dict[str, float] = _compute_scores_from_dists(distance_measures)
        distance_measures: Dict[str, float] = {f"distances.{k}.factor": v for k, v in distance_measures.items()}
    else:
        distance_measures: Dict[str, float] = {}

    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #
    # AGGREGATE DATA - For each linearity measure
    # -~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~- #

    if compute_linearity:
        # stack all into arrays: <shape: (repeats, ...)>
        # then aggregate over first dimension: <shape: (...)>
        # - eg: axis_ratio  (repeats,)        -> ()
        # - eg: axis_values (repeats, z_size) -> (z_size,)
        linear_measures: Dict[str, torch.Tensor] = _torch_stack_all_dicts(linear_measures)
        linear_measures: Dict[str, torch.Tensor] = {k: v.mean(dim=0) for k, v in linear_measures.items()}
        # compute average scores & remove keys
        linear_measures["linearity.axis_ratio_ave.var"] = _score_from_unsorted(
            linear_measures.pop("_TEMP_.axis_values.var"), top_2=False, norm=True
        )  # shape: (z_size,) -> ()
        # convert values
        linear_measures: Dict[str, float] = {k: float(v) for k, v in linear_measures.items()}
    else:
        linear_measures: Dict[str, float] = {}

    # done!
    return {
        **distance_measures,
        **linear_measures,
    }


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


# if __name__ == '__main__':
#
#     def main():
#         import lightning as L
#         from torch.utils.data import DataLoader
#         from disent.frameworks.vae import BetaVae
#         from disent.frameworks.vae import TripletVae
#         from disent.model import AutoEncoder
#         from disent.model.ae import EncoderConv64, DecoderConv64
#         from disent.util.strings import colors
#         from disent.util.profiling import Timer
#         from disent.dataset.data import XYObjectData
#         from disent.dataset.data import XYSquaresData
#         from disent.dataset.sampling import RandomSampler
#         from disent.dataset.sampling import GroundTruthDistSampler
#         from disent.dataset.transform import ToImgTensorF32
#         from disent.nn.weights import init_model_weights
#         from disent.util.seeds import seed
#         import logging
#
#         logging.basicConfig(level=logging.INFO)
#
#         def get_str(r):
#             return ', '.join(f'{k}={v:6.4f}' for k, v in r.items())
#
#         def print_r(name, steps, result, clr=colors.lYLW, t: Timer = None):
#             print(f'{clr}{name:<13} ({steps:>04}){f" {colors.GRY}[{t.pretty}]{clr}" if t else ""}: {get_str(result)}{colors.RST}')
#
#         def calculate(name, steps, dataset, get_repr):
#             with Timer() as t:
#                 r = {
#                     **metric_factored_components.compute_fast(dataset, get_repr),
#                 }
#             results.append((name, steps, r))
#             print_r(name, steps, r, colors.lRED, t=t)
#             print(colors.GRY, '='*100, colors.RST, sep='')
#             return r
#
#         class XYOverlapData(XYSquaresData):
#             def __init__(self, square_size=8, image_size=64, grid_spacing=None, num_squares=3, rgb=True):
#                 if grid_spacing is None:
#                     grid_spacing = (square_size+1) // 2
#                 super().__init__(square_size=square_size, image_size=image_size, grid_spacing=grid_spacing, num_squares=num_squares, rgb=rgb)
#
#         # datasets = [XYObjectData(rgb=False, palette='white'), XYSquaresData(), XYOverlapData(), XYObjectData()]
#         datasets = [XYObjectData()]
#
#         # TODO: fix for dead dimensions
#         # datasets = [XYObjectData(rgb=False, palette='white')]
#
#         results = []
#         for data in datasets:
#             seed(7777)
#
#             # dataset = DisentDataset(data, sampler=RandomSampler(num_samples=1), transform=ToImgTensorF32())
#             # dataloader = DataLoader(dataset=dataset, batch_size=32, shuffle=True, pin_memory=True)
#             # module = init_model_weights(BetaVae(
#             #     model=AutoEncoder(
#             #         encoder=EncoderConv64(x_shape=data.x_shape, z_size=9, z_multiplier=2),
#             #         decoder=DecoderConv64(x_shape=data.x_shape, z_size=9),
#             #     ),
#             #     cfg=BetaVae.cfg(beta=0.0001, loss_reduction='mean', optimizer='adam', optimizer_kwargs=dict(lr=1e-3))
#             # ), mode='xavier_normal')
#
#             dataset = DisentDataset(data, sampler=GroundTruthDistSampler(num_samples=3), transform=ToImgTensorF32())
#             dataloader = DataLoader(dataset=dataset, batch_size=32, shuffle=True, pin_memory=True)
#             module = init_model_weights(TripletVae(
#                 model=AutoEncoder(
#                     encoder=EncoderConv64(x_shape=data.x_shape, z_size=9, z_multiplier=2),
#                     decoder=DecoderConv64(x_shape=data.x_shape, z_size=9),
#                 ),
#                 cfg=TripletVae.cfg(beta=0.0001, loss_reduction='mean', optimizer='adam', optimizer_kwargs=dict(lr=2e-4), triplet_p=1, triplet_loss='triplet_soft', triplet_scale=1)
#             ), mode='xavier_normal')
#
#             gpus = 1 if torch.cuda.is_available() else 0
#
#             # we cannot guarantee which device the representation is on
#             get_repr = lambda x: module.encode(x.to(module.device))
#             # PHASE 1, UNTRAINED
#             L.Trainer(logger=False, enable_checkpointing=False, fast_dev_run=True, gpus=gpus, weights_summary=None).fit(module, dataloader)
#             if torch.cuda.is_available(): module = module.to('cuda')
#             calculate(data.__class__.__name__, 0, dataset, get_repr)
#             # PHASE 2, LITTLE TRAINING
#             L.Trainer(logger=False, enable_checkpointing=False, max_steps=256, gpus=gpus, weights_summary=None).fit(module, dataloader)
#             if torch.cuda.is_available(): module = module.to('cuda')
#             calculate(data.__class__.__name__, 256, dataset, get_repr)
#             # PHASE 3, MORE TRAINING
#             L.Trainer(logger=False, enable_checkpointing=False, max_steps=2048, gpus=gpus, weights_summary=None).fit(module, dataloader)
#             if torch.cuda.is_available(): module = module.to('cuda')
#             calculate(data.__class__.__name__, 256+2048, dataset, get_repr)
#             results.append(None)
#
#         for result in results:
#             if result is None:
#                 print()
#                 continue
#             (name, steps, result) = result
#             print_r(name, steps, result, colors.lYLW)
#
#     main()


# if __name__ == '__main__':
#
#     def _same(ap0, an0, ap1, an1):
#         return ((ap0 < an0) & (ap1 < an1)) | ((ap0 == an0) & (ap1 == an1)) | ((ap0 > an0) & (ap1 > an1))
#
#     def main():
#         ap0 = np.array([1, 1, 1, 1,  1, 1, 1, 1,  2, 2, 2, 2,  2, 2, 2, 2])
#         an0 = np.array([1, 1, 1, 1,  2, 2, 2, 2,  1, 1, 1, 1,  2, 2, 2, 2])
#         ap1 = np.array([1, 1, 2, 2,  1, 1, 2, 2,  1, 1, 2, 2,  1, 1, 2, 2])
#         an1 = np.array([1, 2, 1, 2,  1, 2, 1, 2,  1, 2, 1, 2,  1, 2, 1, 2])
#         print(ap0)
#         print(an0)
#         print(ap1)
#         print(an1)
#
#         print(_same(ap0, an0, ap1, an1).astype('int'))
#
#     main()



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
import pytest
import torch

from disent.dataset.transform import FftGaussianBlur
from disent.dataset.transform._augment import _expand_to_min_max_tuples
from disent.nn.functional import torch_gaussian_kernel
from disent.nn.functional import torch_gaussian_kernel_2d

# ========================================================================= #
# TESTS                                                                     #
# ========================================================================= #


def test_gaussian_kernels():
    r = lambda *shape: torch.stack(torch.meshgrid(*((torch.arange(s) + 1) for s in shape)), dim=-1).max(dim=-1).values

    assert (9,) == torch_gaussian_kernel(sigma=1.0, truncate=4.0).shape
    assert (5, 41) == torch_gaussian_kernel(sigma=r(5), truncate=4.0).shape
    assert (5, 3, 41) == torch_gaussian_kernel(sigma=r(5, 3), truncate=4.0).shape
    assert (5, 7, 71) == torch_gaussian_kernel(sigma=r(5, 1), truncate=r(1, 7)).shape

    assert (9, 7) == torch_gaussian_kernel_2d(sigma=1.0, truncate=4.0, sigma_b=1.0, truncate_b=3.0).shape
    assert (5, 41, 9) == torch_gaussian_kernel_2d(sigma=r(5), truncate=4.0, sigma_b=1.0, truncate_b=4.0).shape
    assert (5, 7, 41, 57) == torch_gaussian_kernel_2d(
        sigma=r(5, 1), truncate=4.0, sigma_b=r(1, 7), truncate_b=4.0
    ).shape


def test_fft_guassian_blur_sigmas():
    assert _expand_to_min_max_tuples(1.0) == ((1.0, 1.0), (1.0, 1.0))
    assert _expand_to_min_max_tuples([0.0, 1.0]) == ((0.0, 1.0), (0.0, 1.0))
    assert _expand_to_min_max_tuples([[0.0, 1.0], [1.0, 2.0]]) == ((0.0, 1.0), (1.0, 2.0))
    with pytest.raises(Exception):
        _expand_to_min_max_tuples([[0.0, 1.0], 1.0])
    with pytest.raises(Exception):
        _expand_to_min_max_tuples([0.0, [1.0, 2.0]])


def test_fft_guassian_blur():
    fn = FftGaussianBlur(sigma=1.0, truncate=3.0)
    fn(torch.randn(256, 3, 64, 64))


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import warnings
from typing import Tuple

from torch import Tensor
from torch import nn

from disent.model import DisentDecoder
from disent.model import DisentEncoder
from disent.nn.activations import Swish

# ========================================================================= #
# disentanglement_lib Conv models                                           #
# ========================================================================= #


class EncoderConv64Norm(DisentEncoder):
    """
    Modified version of `EncoderConv64` with
    selectable activations and norm layers
    """

    def __init__(
        self, x_shape=(3, 64, 64), z_size=6, z_multiplier=1, activation="leaky_relu", norm="layer", norm_pre_act=True
    ):
        # checks
        (C, H, W) = x_shape
        assert (H, W) == (64, 64), "This model only works with image size 64x64."
        super().__init__(x_shape=x_shape, z_size=z_size, z_multiplier=z_multiplier)

        # warnings
        warnings.warn(f"this model: {self.__class__.__name__} is non-standard in VAE research!")

        self.model = nn.Sequential(
            nn.Conv2d(in_channels=C, out_channels=32, kernel_size=4, stride=2, padding=2),
            *_make_activations(activation=activation, norm=norm, shape=(32, 33, 33), norm_pre_act=norm_pre_act),
            nn.Conv2d(in_channels=32, out_channels=32, kernel_size=4, stride=2, padding=2),
            *_make_activations(activation=activation, norm=norm, shape=(32, 17, 17), norm_pre_act=norm_pre_act),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=2, stride=2, padding=1),
            *_make_activations(activation=activation, norm=norm, shape=(64, 9, 9), norm_pre_act=norm_pre_act),
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=2, stride=2, padding=1),
            *_make_activations(activation=activation, norm=norm, shape=(64, 5, 5), norm_pre_act=norm_pre_act),
            nn.Flatten(),
            nn.Linear(in_features=1600, out_features=256),
            *_make_activations(activation=activation, norm="none"),
            nn.Linear(in_features=256, out_features=self.z_total),
        )

    def encode(self, x) -> (Tensor, Tensor):
        return self.model(x)


class DecoderConv64Norm(DisentDecoder):
    """
    Modified version of `DecoderConv64` with
    selectable activations and norm layers
    """

    def __init__(
        self, x_shape=(3, 64, 64), z_size=6, z_multiplier=1, activation="leaky_relu", norm="layer", norm_pre_act=True
    ):
        # checks
        (C, H, W) = x_shape
        assert (H, W) == (64, 64), "This model only works with image size 64x64."
        super().__init__(x_shape=x_shape, z_size=z_size, z_multiplier=z_multiplier)

        # warnings
        warnings.warn(f"this model: {self.__class__.__name__} is non-standard in VAE research!")

        self.model = nn.Sequential(
            nn.Linear(in_features=self.z_size, out_features=256),
            *_make_activations(activation=activation, norm="none"),
            nn.Linear(in_features=256, out_features=1024),
            *_make_activations(activation=activation, norm="none"),
            nn.Unflatten(dim=1, unflattened_size=[64, 4, 4]),
            nn.ConvTranspose2d(in_channels=64, out_channels=64, kernel_size=4, stride=2, padding=1),
            *_make_activations(activation=activation, norm=norm, shape=(64, 8, 8), norm_pre_act=norm_pre_act),
            nn.ConvTranspose2d(in_channels=64, out_channels=32, kernel_size=4, stride=2, padding=1),
            *_make_activations(activation=activation, norm=norm, shape=(32, 16, 16), norm_pre_act=norm_pre_act),
            nn.ConvTranspose2d(in_channels=32, out_channels=32, kernel_size=4, stride=2, padding=1),
            *_make_activations(activation=activation, norm=norm, shape=(32, 32, 32), norm_pre_act=norm_pre_act),
            nn.ConvTranspose2d(in_channels=32, out_channels=C, kernel_size=4, stride=2, padding=1),
        )

    def decode(self, z) -> Tensor:
        return self.model(z)


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


def _make_activations(activation="relu", inplace=True, norm="layer", shape: Tuple[int, ...] = None, norm_pre_act=True):
    # get activation layer
    if activation == "relu":
        a_layer = nn.ReLU(inplace=inplace)
    elif activation == "leaky_relu":
        a_layer = nn.LeakyReLU(inplace=inplace)
    elif activation == "swish":
        a_layer = Swish()
    else:
        raise KeyError(f"invalid activation layer: {repr(activation)}")

    # get norm layer
    # https://www.programmersought.com/article/41731094913/
    # TODO: nn.GroupNorm

    if norm in (None, "none"):
        n_layer = None
    else:
        C, H, W = shape
        if norm == "batch":
            n_layer = nn.BatchNorm2d(num_features=C)
        elif norm == "instance":
            n_layer = nn.InstanceNorm2d(num_features=C)
        elif norm == "layer":
            n_layer = nn.LayerNorm(normalized_shape=[H, W])
        elif norm == "layer_chn":
            n_layer = nn.LayerNorm(normalized_shape=[C, H, W])
        else:
            raise KeyError(f"invalid norm layer: {repr(norm)}")
    # order layers
    layers = (n_layer, a_layer) if norm_pre_act else (a_layer, n_layer)
    # return layers
    return tuple(l for l in layers if l is not None)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import pytest
import torch

from disent.dataset import DisentDataset
from disent.dataset.data import XYObjectData
from disent.dataset.transform import ToImgTensorF32
from disent.metrics import *
from disent.util.function import wrapped_partial

# ========================================================================= #
# TESTS                                                                     #
# ========================================================================= #


@pytest.mark.parametrize(
    "metric_fn",
    [
        wrapped_partial(metric_mig, num_train=7),
        wrapped_partial(metric_unsupervised, num_train=7),
        wrapped_partial(metric_dci, num_train=7, num_test=7),
        wrapped_partial(metric_sap, num_train=7, num_test=7),
        wrapped_partial(metric_factor_vae, num_train=7, num_eval=7, num_variance_estimate=7),
        wrapped_partial(metric_flatness, repeats=7),
        wrapped_partial(metric_factored_components, repeats=7),
    ],
)
def test_metrics(metric_fn):
    z_size = 8
    # ground truth data
    # TODO: DisentDataset should not be needed to compute metrics!
    dataset = DisentDataset(XYObjectData(), transform=ToImgTensorF32())
    # randomly sampled representation
    get_repr = lambda x: torch.randn(len(x), z_size)
    # evaluate
    metric_fn(dataset, get_repr)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from functools import wraps
from typing import Callable
from typing import Dict
from typing import NoReturn
from typing import Optional
from typing import Union

from disent.util.inout.hashing import hash_file
from disent.util.inout.hashing import normalise_hash
from disent.util.inout.hashing import validate_file_hash

log = logging.getLogger(__name__)


# ========================================================================= #
# Function Caching                                                          #
# ========================================================================= #


class stalefile(object):
    """
    decorator that only runs the wrapped function if a
    file does not exist, or its hash does not match.
    """

    def __init__(
        self,
        file: str,
        hash: Optional[Union[str, Dict[str, str]]],
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        self.file = file
        self.hash = normalise_hash(hash=hash, hash_mode=hash_mode)
        self.hash_type = hash_type
        self.hash_mode = hash_mode

    def __call__(self, func: Callable[[str], NoReturn]) -> Callable[[], str]:
        @wraps(func)
        def wrapper() -> str:
            if self.is_stale():
                log.debug(f"calling wrapped function: {func} because the file is stale: {repr(self.file)}")
                func(self.file)
                validate_file_hash(
                    self.file, hash=self.hash, hash_type=self.hash_type, hash_mode=self.hash_mode, missing_ok=True
                )
            else:
                log.debug(f"skipped wrapped function: {func} because the file is fresh: {repr(self.file)}")
            return self.file

        return wrapper

    def is_stale(self):
        fhash = hash_file(file=self.file, hash_type=self.hash_type, hash_mode=self.hash_mode, missing_ok=True)
        if not fhash:
            log.info(f"file is stale because it does not exist: {repr(self.file)}")
            return True
        if fhash != self.hash:
            log.warning(
                f"file is stale because the computed {self.hash_mode} {self.hash_type} hash: {fhash} does not match the target hash: {self.hash} for file: {repr(self.file)}"
            )
            return True
        log.debug(f"file is fresh: {repr(self.file)}")
        return False

    def __bool__(self):
        return self.is_stale()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from dataclasses import dataclass
from numbers import Number
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple
from typing import Union

import torch

from disent.frameworks.ae._supervised__tae import TripletAe
from disent.frameworks.ae._weaklysupervised__adaae import AdaAe
from disent.frameworks.vae._supervised__adaneg_tvae import AdaNegTripletVae

log = logging.getLogger(__name__)


# ========================================================================= #
# Guided Ada Vae                                                            #
# ========================================================================= #


class AdaNegTripletAe(TripletAe):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    This is the supervised version of the Adaptive Triplet Loss for Auto-Encoders not VAEs
    - Triplets are usually ordered using the ground-truth distances over observations
    - Adaptive Triplet Loss is used to guide distance learning and encourage disentanglement
    """

    REQUIRED_OBS = 3

    @dataclass
    class cfg(TripletAe.cfg, AdaAe.cfg):
        # ada_tvae - loss
        adat_triplet_share_scale: float = 0.95

    def hook_ae_compute_ave_aug_loss(
        self, zs: Sequence[torch.Tensor], xs_partial_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]
    ) -> Tuple[Union[torch.Tensor, Number], Dict[str, Any]]:
        return AdaNegTripletVae.estimate_ada_triplet_loss_from_zs(
            zs=zs,
            cfg=self.cfg,
        )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
